! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: especplot_mod
!
! Authors:  Didier Barret - Jean Francois Olive  - CESR
!          
! Description: Plot background and source spectra as measured by epic
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module especplot_mod

  use dal
  use cal
  use param  
  use ErrorHandling

  implicit none 
  
  private 

  public especplot

contains 
  
  !-------------------------------------------------------------------------------------
  subroutine especplot
  !-------------------------------------------------------------------------------------
  ! Input parameters :
  !   Src_File_Name , Bkg_File_Name : Source (Src) and background (Bkg) FITS file name
  !   Device_Name : Plot device type
  !   Output_File : Output file name
  !   Binning_Auto , Binning_Size : Binning specifications (Automatimatic/Manual & size)
     
    character(len=1024) :: Src_File_Name, Bkg_File_Name
    character(len=1024) :: Device_Name , Output_File
    character(len=3) :: Binning_Auto
    character(len=3) :: Bin_Channels
    real :: Binning_Size
    character(len=*),parameter :: cifName = ''

    ! Channel : Vector containing the PI channels
    ! Emin , Emax , Emed , Edelta : Energy boundaries, medium and width of the bins
    ! Src_Counts , Bkg_Counts , Net_Counts : Counts spectrum (Net=Src-normalized(Bkg))
    ! Net_Counts_Err , Bkg_Counts_Err : Errors associated with counts
    ! Size_Vectors : Automatically updated size of all these vectors

    integer(kind=int16),dimension(:),pointer :: Channel
    real(kind=single),dimension(:),pointer :: Emin , Emax , Emed , Edelta
    real(kind=single),dimension(:),pointer :: Src_Counts , Bkg_Counts
    real(kind=single),dimension(:),pointer :: Net_Counts , Net_Counts_Err
    real(kind=single),dimension(:),pointer :: Bkg_Counts_Err
    integer :: Size_Vectors

    real :: Src_Exposure_Time , Bkg_Exposure_Time
    character(len=5) :: Instrument

    ! Plot variables : Viewports boundaries(Window1&2) and labels

    real,dimension(4) :: Window1
    real,dimension(4) :: Window2
    character(len=1000) :: Label_Gen , Label_Obs , Label_Exposure , Label_Time

    Src_File_Name = StringParameter("srcspectrumset")
    Bkg_File_Name = StringParameter("bkgspectrumset")
    Binning_Auto = 'YES'    
    Bin_Channels = 'YES'   
    If (BooleanParameter("withbinning")) Binning_Auto='NO'
    If (BooleanParameter("mincounts")) Bin_Channels='NO'   
    If (BooleanParameter("mincounts") .and. BooleanParameter("channels"))  then
      call warning('ESPECPLOT','Only grouping by channels OR minimum counts allowed : Set to default channels') 
      Bin_Channels = 'YES' 
    end if
    Binning_Size = RealParameter("binsize")
    Device_Name = change_case(StringParameter("plotdevice"),2)
    Output_File = StringParameter("plotfile")

    call read_file (Src_File_Name, Bkg_File_Name, Instrument, &
      Channel, Src_Counts, Bkg_Counts, Src_Exposure_Time, Bkg_Exposure_Time, &
      Label_Gen, Label_Obs, Label_Exposure, Label_Time, Emin, Emax, Size_Vectors)

    call open_plot(Device_Name, Output_File, &
         Label_Gen, Label_Obs, Label_Exposure, Label_Time, Window1, Window2)

    call check_binning_input_data(Binning_Size)

    allocate(Emed(Size_Vectors), Edelta(Size_Vectors))
    Emed=(Emin+Emax)/2.0
    Edelta=(Emax-Emin)/2.0
    allocate(Net_Counts(Size_Vectors), Net_Counts_Err(Size_Vectors))
    Net_Counts = Src_Counts - Bkg_Counts*Src_Exposure_Time/Bkg_Exposure_Time
    Net_Counts_Err = sqrt(Src_Counts + Bkg_Counts* &
      ((Src_Exposure_Time / Bkg_Exposure_Time)**2))
    deallocate(Src_Counts)

    call rebin(Binning_Auto, Bin_Channels, Binning_Size, Size_Vectors, Emed, Edelta, &
      Net_Counts, Net_Counts_Err)
    call normalize_spectrum(Edelta, Net_Counts, Net_Counts_Err, Src_Exposure_Time)

    call eliminate_bad_values(Size_Vectors, Emed, Edelta, Net_Counts, Net_Counts_Err)

    call draw_plot(Window1, 'EPIC SOURCE SPECTRUM', Size_Vectors, &
         Emed, Edelta, Net_Counts, Net_Counts_Err)

    Size_Vectors = Size(Emin)
    deallocate(Emed, Edelta, Net_Counts, Net_Counts_Err)
    allocate(Emed(Size_Vectors), Edelta(Size_Vectors))
    Emed=(Emin+Emax)/2.0
    Edelta=(Emax-Emin)/2.0
    allocate(Bkg_Counts_Err(Size_Vectors))
    Bkg_Counts_Err=sqrt(Bkg_Counts)

    call rebin(Binning_Auto, Bin_Channels, Binning_Size, Size_Vectors, Emed, Edelta, &
      Bkg_Counts, Bkg_Counts_Err)

    call normalize_spectrum(Edelta, Bkg_Counts, Bkg_Counts_Err, Bkg_Exposure_Time)

    call eliminate_bad_values(Size_Vectors, Emed, Edelta, Bkg_Counts, Bkg_Counts_Err)

    call draw_plot(Window2, 'EPIC BACKGROUND SPECTRUM', Size_Vectors, &
         Emed, Edelta, Bkg_Counts, Bkg_Counts_Err)

    deallocate(Emed, Edelta, Bkg_Counts, Bkg_Counts_Err)

    call pgend

  end subroutine especplot

  !-------------------------------------------------------------------------------------
  subroutine read_file(Src_File_Name, Bkg_File_Name, Instrument, Channel, &
    Src_Counts, Bkg_Counts, Src_Exposure_Time, Bkg_Exposure_Time, &
    Label_Gen, Label_Obs, Label_Exposure, Label_Time, Emin, Emax, Size_Vectors)
  !-------------------------------------------------------------------------------------
  ! Reads Src&Bkg FITS file and recovers all data needed 

    character(len=*),intent(in) :: Src_File_Name , Bkg_File_Name 
    character(len=*),intent(out) :: Label_Gen , Label_Obs
    character(len=*) :: Instrument
    character(len=*),intent(out) :: Label_Exposure , Label_Time
    real,intent(out) :: Src_Exposure_Time , Bkg_Exposure_Time
    real :: Chanval, Chanvalback
    integer :: Binfactor, Binfactorback
    type( DataSetT ) :: Src_Set , Bkg_Set
    type( TableT ) :: Src_Table , Bkg_Table
    type( ColumnT ) :: Src_Col_Channel , Src_Col_Counts
    type( ColumnT ) :: Bkg_Col_Channel , Bkg_Col_Counts
    integer(kind=int16),dimension(:),pointer :: Channel , Bkg_Channel
    integer(kind=int32),dimension(:),pointer :: Int_Src_Counts , Int_Bkg_Counts
    real(kind=single),dimension(:),pointer :: Src_Counts , Bkg_Counts
    integer:: Size_Vectors
    real(kind=single),dimension(:),pointer  :: Emin , Emax
    character(len=*),parameter :: cifName = ''
 
    call read_data (Src_File_Name,Src_Set)
    call read_data (Bkg_File_Name,Bkg_Set)
    call read_table (Src_File_Name, Src_Set, Src_Table, Src_Exposure_Time, & 
                                                                      Chanval, Binfactor)
    call read_table (Bkg_File_Name, Bkg_Set, Bkg_Table, Bkg_Exposure_Time, & 
                                                              Chanvalback, Binfactorback)
    Src_Col_Channel=Column(Src_Table, ColumnNumber(Src_Table,'CHANNEL'),Read)
    Bkg_Col_Channel=Column(Bkg_Table, ColumnNumber(Bkg_Table,'CHANNEL'),Read)
    Src_Col_Counts=Column(Src_Table, ColumnNumber(Src_Table,'COUNTS'),Read)
    Bkg_Col_Counts=Column(Bkg_Table, ColumnNumber(Bkg_Table,'COUNTS'),Read)
    Channel => int16Data(Src_Col_Channel)
    Bkg_Channel => int16Data(Bkg_Col_Channel)
    Int_Src_Counts => int32Data(Src_Col_Counts)
    Int_Bkg_Counts => int32Data(Bkg_Col_Counts)
    allocate( Src_Counts(size(Int_Src_Counts)) )
    allocate( Bkg_Counts(size(Int_Bkg_Counts)) )
    Src_Counts = 0.0
    Src_Counts = real(Int_Src_Counts)
    Bkg_Counts = real(Int_Bkg_Counts)
    call check_Src_Bkg_coherence (Src_Set,Bkg_Set,size(Channel),size(Bkg_Channel))
    Instrument = change_case(StringAttribute(Src_Set,'INSTRUME'),2)
    call create_labels (Src_Set,Src_Table,Bkg_Table,Label_Gen,Label_Obs,& 
                                                         Label_Exposure,Label_Time)

    call CAL_operations(Instrument, Channel, Binfactor, Emin, Emax, Size_Vectors)

    call release(Src_Col_Channel)
    call release(Src_Col_Counts)
    call release(Src_Table)    
    call release(Src_Set) 
    call release(Bkg_Col_Channel)
    call release(Bkg_Col_Counts)
    call release(Bkg_Table)
    call release(Bkg_Set)

  end subroutine read_file

  !-------------------------------------------------------------------------------------
  subroutine read_data (File_Name, Set)
  !-------------------------------------------------------------------------------------
  ! Reads the in parameter FITS file and gives back a handle to corresponding dataset.
  ! Checks some keywords existence and validity.
  ! Also gives back content title and exposure time

    character(len=*),intent(in) :: File_Name 
    type( DataSetT ),intent(out) :: Set

    Set = dataSet(File_Name,READ)
    if ( .not. HasAttribute(Set,"INSTRUME") ) &
      call error('ESPECPLOT','Keyword INSTRUMENT missing in the input file '//File_Name)

!    if ( .not. HasAttribute(Set,"CONTENT") ) &
!      call error('ESPECPLOT','Keyword CONTENT missing in the input file '//File_Name)
!
!    if ( change_case(StringAttribute(Set,"CONTENT"),2) /= Content ) &
!      call error('ESPECPLOT','Invalid CONTENT parameter for input file '//File_Name)

! According to SPR 1119; the existence of this keyword shouldn't be checked

  end subroutine read_data

  !-------------------------------------------------------------------------------------
  subroutine read_table (File_Name, Set, Tab, Exposure, Chanval, Binfactor)
  !-------------------------------------------------------------------------------------
  ! With FITS file name and handle to dataset, generates table.
  ! Checks more keywords and reads the binning factor and average channel value

    character(len=*),intent(in) :: File_name
    type( DataSetT ),intent(in) :: Set
    type( TableT ),intent(out) :: Tab
    real,intent(out) :: Exposure, Chanval
    integer :: Binfactor

    Tab = table(Set,BlockNumber(Set,'SPECTRUM'))
    if ( .not. HasAttribute(Tab,"HDUCLASS") ) &
      call error('ESPECPLOT','Keyword HDUCLAS missing in the input file '//File_Name)
    if ( change_case(StringAttribute(Tab,"HDUCLASS"),2) /= 'OGIP' ) &
      call error('ESPECPLOT','HDUCLASS is not equal to OGIP')
    if ( .not. HasAttribute(Tab,"HDUCLAS1") ) &
      call error('ESPECPLOT','Keyword HDUCLAS1 missing in the input file '//File_Name)
    if ( change_case(StringAttribute(Tab,"HDUCLAS1"),2) /= 'SPECTRUM' ) &
      call error('ESPECPLOT','HDUCLAS1 is not equal to SPECTRUM')
    if ( .not. HasAttribute(Tab,"HDUCLAS2") ) &
      call error('ESPECPLOT','Keyword HDUCLAS2 missing in the input file '//File_Name)
    if ( change_case(StringAttribute(Tab,"HDUCLAS2"),2) /= 'TOTAL' ) &
      call error('ESPECPLOT','HDUCLAS2 is not equal to TOTAL')
    if ( .not. HasAttribute(Tab,"HDUCLAS3")) &
      call error('ESPECPLOT','Keyword HDUCLAS3 missing in the input file '//File_Name)
    if ( change_case(StringAttribute(Tab,"HDUCLAS3"),2) /= 'COUNT' ) &
      call error('ESPECPLOT','HDUCLAS3 is not equal to COUNT')
    if ( .not. HasAttribute(Tab,"CHANTYPE") ) &
      call error('ESPECPLOT','Keyword CHANTYPE missing in the input file '//File_Name)
    if ( change_case(StringAttribute(Tab,"CHANTYPE"),2) /= 'PI' ) &
      call error('ESPECPLOT','CHANTYPE is not equal to PI')

    if ( .not. HasAttribute(Tab,"EXPOSURE") ) &
      call error('ESPECPLOT','Keyword EXPOSURE missing in the input file '//File_Name)
    Exposure=real32Attribute(Tab,"EXPOSURE")
    if ( Exposure < epsilon(0.0) ) call error('ESPECPLOT', &
      'Null or Negative exposure time in the input file '//File_Name) 
    if ( .not. HasAttribute(Tab,"SPECDELT") ) &
      call error('ESPECPLOT','Keyword SPECDELT missing in the input file '//File_Name)
    Binfactor=int16Attribute(Tab,"SPECDELT")
    if ( Binfactor < epsilon(0.0) ) call error('ESPECPLOT', &
      'Null or Negative binning factor in the input file '//File_Name) 
    if ( .not. HasAttribute(Tab,"SPECVAL") ) &
      call error('ESPECPLOT','Keyword SPECVAL missing in the input file '//File_Name)
    Chanval=real32Attribute(Tab,"SPECVAL")
    if ( Exposure < epsilon(0.0) ) call error('ESPECPLOT', &
      'Null or Negative channel value in the input file '//File_Name) 

  end subroutine read_table

  !-------------------------------------------------------------------------------------
  subroutine check_Src_Bkg_coherence (Src_Set, Bkg_Set, Src_Ch_Size, Bkg_Ch_Size)
  !-------------------------------------------------------------------------------------
    type( dataSetT ),intent(in) :: Src_set, Bkg_Set
    integer,intent(in) :: Src_Ch_Size, Bkg_Ch_Size

    if ( change_case(StringAttribute(Bkg_Set,"INSTRUME"),2) /= &
      change_case(StringAttribute(Src_Set,"INSTRUME"),2) ) call error('ESPECPLOT', &
        'Not the same instrument for the source and backgroud files')
    if ( Src_Ch_Size /= Bkg_Ch_Size ) &
      call error('ESPECPLOT', &
        'PI Channels are different for the source and background spectra')

  end subroutine check_Src_Bkg_coherence

  !-------------------------------------------------------------------------------------
  subroutine create_labels(Src_Set, Src_Table, Bkg_Table, &
    Label_Gen, Label_Obs, Label_Exposure, Label_Time)
  !-------------------------------------------------------------------------------------
    type( dataSetT ),intent(in) :: Src_set 
    type( TableT ),intent(in)  :: Src_Table, Bkg_Table
    character(len=*),intent(out) :: Label_Gen,Label_Obs,Label_Exposure,Label_Time

    Label_Gen = trim(StringAttribute(Src_Set,'TELESCOP'))//' - '// &
                trim(StringAttribute(Src_Set,'INSTRUME'))//' - '// &
                trim(StringAttribute(Src_Set,'DATAMODE'))//' - '//'PI Name : '// &
                trim(StringAttribute(Src_Set,'OBSERVER'))
    Label_Obs = 'OBJECT : '//trim(StringAttribute(Src_Set,'OBJECT'))//' - '// &
                'OBS_ID : '//trim(StringAttribute(Src_Set,'OBS_ID'))
    Label_Time = 'START : '//trim(StringAttribute(Src_Set,'DATE-OBS'))//' - '// &
                 'STOP : '//trim(StringAttribute(Src_Set,'DATE-END'))
    Label_Exposure = 'SOURCE EXPOSURE TIME='//trim(StringAttribute(Src_Table, &
      'EXPOSURE'))//' sec - BACKGROUND EXPOSURE TIME='//trim(StringAttribute(Bkg_Table, &
      'EXPOSURE'))//' sec'

  end subroutine create_labels

  !-------------------------------------------------------------------------------------
  subroutine CAL_operations(Instrument, Channel, Binfactor, Emin, Emax, Size_Vectors)
  !-------------------------------------------------------------------------------------
  ! Gets energy boundaries  - changed from using the CAL.  The boundaries are now 
  !calculated using the binning factor and 1PI=1eV, a fundamental convention adopted 
  !in the pipeline. 

    character(len=*),intent(in) :: Instrument
    integer(kind=int16),dimension(:), intent(in) :: Channel
    integer,intent(out) :: Size_Vectors
    real(kind=single),dimension(:),pointer :: Emin , Emax
    integer::Binfactor, i

!    call CAL_openCCF(cifName)
    select case (Instrument)
      case ('EMOS1')
        call CAL_setState(EMOS1)
      case ('EMOS2')
        call CAL_setState(EMOS2)
      case ('EPN')
        call CAL_setState(EPN)
      case default
        call error('ESPECPLOT', &
             'Incorrect instrument name : should be either EMOS1, EMOS2, EPN')
    end select
!    call CAL_getEbounds(Channel , Emin_Dbl , Emax_Dbl)
!    call CAL_getEbins(Channel , Emin_Dbl , Emax_Dbl)
    Size_Vectors=size(Channel)
    allocate(Emin(Size_Vectors))
    allocate(Emax(Size_Vectors))
!    Emin=sngl(Emin_dbl)/1000.
!    Emax= sngl(Emax_dbl)/1000.
 
     Emin = 0.0; Emax = 0.0

     do i=1,Size_Vectors-1
       Emin(i+1)=Emin(i)+Binfactor
       Emax(i)=Emin(i+1)
     end do

     Emax(Size_Vectors)=Emax(Size_Vectors-1)+Binfactor

!    call CAL_ReleaseMemoryReal2OneDim(Emin_Dbl , Emax_Dbl)

  end subroutine CAL_operations

  !-------------------------------------------------------------------------------------
  subroutine check_binning_input_data(Binning_Size)
  !-------------------------------------------------------------------------------------

    real,intent(inout) :: Binning_Size

    if ( Binning_Size<-999. ) then
      call warning('ESPECPLOT','Invalid binning size : Set to default=1.0') 
      Binning_Size=1.
    end if

  end subroutine check_binning_input_data

  !-------------------------------------------------------------------------------------
  subroutine rebin(Binning_Auto,Bin_Channels,Binning_Size,Size_Vectors,Emed,Edelta,Counts,Counts_Err)
  !-------------------------------------------------------------------------------------
  ! Rebins Emed , Edelta , Counts , Counts_Err.
  ! In case of automatic rebinning, the stop condition is either number of bins reached
  ! lower than 25 or all rebinned counts are larger than 10.

    character(len=3),intent(in) :: Binning_Auto, Bin_Channels
    real,intent(in) :: Binning_Size
    integer,intent(inout) :: Size_Vectors
    real(kind=single),dimension(:),pointer :: Emed , Edelta
    real(kind=single),dimension(:),pointer :: Counts , Counts_Err

    ! Emed_Reb , Edelta_Reb , Counts_Reb , Counts_Err_Reb : Temporary vectors containing
    ! energies and counts rebinned
    ! Binning_Scheme : Temporary vector containing rebinning information

    real(kind=single),dimension(:),pointer :: Emed_Reb, Edelta_Reb
    real(kind=single),dimension(:),pointer :: Counts_Reb , Counts_Err_Reb
    integer,dimension(:),pointer :: Binning_Scheme
    integer :: Size_Rebin_Vectors
    real :: Binning_Size_Test,groupvalue
    integer :: i,j,tally
    logical :: stop_condition 

    nullify(Emed_Reb) ; nullify(Edelta_Reb)
    nullify(Counts_Reb) ; nullify(Counts_Err_Reb)
    if ( change_case(Binning_Auto,2) == 'YES' ) then
      
      ! Automatic rebinning

      stop_condition = .false.
      Binning_Size_Test = -1.0
      do while ( .not. stop_condition)
        Binning_Size_Test = Binning_Size_Test - 0.1
        call compute_binning_scheme(Bin_Channels, Size_Vectors,Binning_Size_Test, &
	  Size_Rebin_Vectors,Binning_Scheme)
        call make_rebin (Size_Rebin_Vectors,Binning_Scheme,Emed,Edelta, &
          Counts,Counts_Err,Emed_Reb,Edelta_Reb,Counts_Reb,Counts_Err_Reb)
        stop_condition = .true.
        do i = 1, Size_Rebin_Vectors - 1
          if (Counts_Reb(i) < 10.0)  stop_condition = .false.
        end do
        if ( Size_Rebin_Vectors <= 25 )  stop_condition = .true.
      end do
    else if (change_case(Binning_Auto,2) == 'NO' .and. change_case(Bin_Channels,2) == 'YES') then   
      ! Manual rebinning

      call compute_binning_scheme(Bin_Channels, Size_Vectors,Binning_Size , &
          Size_Rebin_Vectors,Binning_Scheme)

      call make_rebin (Size_Rebin_Vectors,Binning_Scheme,Emed,Edelta, &
          Counts,Counts_Err,Emed_Reb,Edelta_Reb,Counts_Reb,Counts_Err_Reb)
    else if (change_case(Binning_Auto,2) == 'NO' .and. change_case(Bin_Channels,2) == 'NO') then 
      j=1
      do i = 1, Size_Vectors
        groupvalue = groupvalue + Counts(i)
        if (groupvalue > Binning_Size) then
          j=j+1
          groupvalue = 0.0
        end if 
      end do

      Size_Rebin_Vectors = j-1
      allocate(Binning_Scheme(Size_Rebin_Vectors))
      Binning_Scheme = 1

      j=1; i=0; tally=0
      do j = 1, Size_Rebin_Vectors-1
        tally = 0
        groupvalue = 0.0
          do while (groupvalue < Binning_Size)
            i=i+1
            groupvalue = groupvalue + Counts(i)
            tally = tally + 1
            Binning_Scheme(j) = tally
        end do
      end do

      call make_rebin (Size_Rebin_Vectors,Binning_Scheme,Emed,Edelta, &
          Counts,Counts_Err,Emed_Reb,Edelta_Reb,Counts_Reb,Counts_Err_Reb)
    end if
    deallocate(Emed , Edelta , Counts , Counts_Err)
    allocate(Emed(Size_Rebin_Vectors) , Edelta(Size_Rebin_Vectors) )
    allocate(Counts(Size_Rebin_Vectors) , Counts_Err(Size_Rebin_Vectors) )
    Emed = Emed_Reb ; Edelta = Edelta_Reb
    Counts = Counts_Reb ; Counts_Err = Counts_Err_Reb
    deallocate(Binning_Scheme , Emed_Reb , Edelta_Reb , Counts_Reb , Counts_Err_Reb)
    Size_Vectors=Size_Rebin_Vectors
  end subroutine rebin

  !-------------------------------------------------------------------------------------
  subroutine compute_binning_scheme(Bin_Channels, Size_Vectors, Binning_Size, &
          Size_Rebin_Vectors , Binning_Scheme)
  !-------------------------------------------------------------------------------------
  ! Defines the rebinning mode and generates Binning_Scheme in function of Binning_Size.

    character(len=3),intent(in) :: Bin_Channels
    integer,intent(in) :: Size_Vectors 
    real,intent(in) :: Binning_Size 
    integer,intent(out) :: Size_Rebin_Vectors
    integer,dimension(:),pointer :: Binning_Scheme

    integer,dimension(Size_Vectors) :: Bin_Numbers
    real :: Abs_Binning_Size , coeff
    integer :: Int_Binning_Size , i

    Size_Rebin_Vectors = 0
   if (Binning_Size> 1. ) then
      
      ! Linear rebinning
      
      Size_Rebin_Vectors = (Size_Vectors-1)/int(Binning_Size) + 1
      allocate(Binning_Scheme(Size_Rebin_Vectors))
      Int_Binning_Size = int(Binning_Size)
      do i = 1,Size_Rebin_Vectors-1
        Binning_Scheme(i) = Int_Binning_Size
      end do
      Binning_Scheme(Size_Rebin_Vectors) = Size_Vectors - &
        (Size_Rebin_Vectors-1)*Int_Binning_Size
    else if (change_case(Bin_Channels,2) == 'YES' .and. Binning_Size<-1. ) then
      
      ! Geometrical  rebinning
      
      Abs_Binning_Size = abs(Binning_Size)
      coeff = Abs_Binning_Size
      i = int(coeff)
      do while ( i < Size_Vectors )
        Size_Rebin_Vectors = Size_Rebin_Vectors + 1
        Bin_Numbers(Size_Rebin_Vectors) = int(coeff)
	coeff = coeff * Abs_Binning_Size
	i = i + int(coeff)
      end do
      Size_Rebin_Vectors = Size_Rebin_Vectors + 1
      Bin_Numbers(Size_Rebin_Vectors) = Size_Vectors - i + int(coeff)
      allocate(Binning_Scheme(Size_Rebin_Vectors))
      Binning_Scheme = Bin_Numbers
!    else if (change_case(Bin_Channels,2) == 'NO' .and. Binning_Size>=1. ) then

!Need to add something in here to make it rebin by counts and not channels

    else
      
      ! No rebinning
      
      Size_Rebin_Vectors = Size_Vectors
      allocate(Binning_Scheme(Size_Rebin_Vectors))
      do i=1,Size_Rebin_Vectors
        Binning_Scheme(i)=1
      end do
    endif

  end subroutine compute_binning_scheme

  !-------------------------------------------------------------------------------------
  subroutine make_rebin (Size_Rebin_Vectors,Binning_Scheme,Emed,Edelta, &
          Counts,Counts_Err,Emed_Reb,Edelta_Reb,Counts_Reb,Counts_Err_Reb)
  !-------------------------------------------------------------------------------------
  ! In parameters (Emed , Edelta , Counts , Counts_Err) are rebinned following
  ! Binning_Scheme. Result vectors Emed_Reb , ... , Counts_Err_Reb have to be reset as
  ! make_rebin can be called several times while automatic binning
  
    integer,intent(in) :: Size_Rebin_Vectors
    real(kind=single),dimension(:),pointer :: Emed , Edelta
    real(kind=single),dimension(:),pointer :: Counts , Counts_Err
    real(kind=single),dimension(:),pointer :: Emed_Reb, Edelta_Reb
    real(kind=single),dimension(:),pointer :: Counts_Reb , Counts_Err_Reb
    integer,dimension(:),pointer :: Binning_Scheme

    integer ::  i , j , k
    real(kind=single) :: Emed_Sum , Edelta_Sum , Counts_Sum , Counts_Err_Sum

    if ( associated(Emed_Reb) ) deallocate(Emed_Reb)
    if ( associated(Edelta_Reb) ) deallocate(Edelta_Reb)
    if ( associated(Counts_Reb) ) deallocate(Counts_Reb)
    if ( associated(Counts_Err_Reb) ) deallocate(Counts_Err_Reb)
    allocate(Emed_Reb(Size_Rebin_Vectors) , Edelta_Reb(Size_Rebin_Vectors) )
    allocate(Counts_Reb(Size_Rebin_Vectors) , Counts_Err_Reb(Size_Rebin_Vectors) )
    k = 0 
    do i = 1,Size_Rebin_Vectors
      Emed_Sum  = 0.0
      Edelta_Sum  = 0.0
      Counts_Sum = 0.0
      Counts_Err_Sum = 0.0
      do j = 1,Binning_Scheme(i)
        k = k + 1
        Emed_Sum  = Emed_Sum + Emed(k)
        Edelta_Sum = Edelta_Sum + Edelta(k)
        Counts_Sum  = Counts_Sum + Counts(k)
        Counts_Err_Sum  = Counts_Err_Sum + Counts_Err(k)**(2.0)
      end do
      Emed_Reb(i)  = Emed_Sum/real(Binning_Scheme(i))
      Edelta_Reb(i)  = Edelta_Sum
      Counts_Reb(i) = Counts_Sum
      Counts_Err_Reb(i) = sqrt(Counts_Err_Sum)
    end do
       
  end subroutine make_rebin

  !-------------------------------------------------------------------------------------
  subroutine eliminate_bad_values(Size_Rebin_Vectors , &
    Emed_Reb , Edelta_Reb , Counts_Reb , Counts_Err_Reb)
  !-------------------------------------------------------------------------------------
  ! Eliminates values which cannot be plotted logarithmically

    real(kind=single),dimension(:),pointer :: Emed_Reb , Edelta_Reb
    real(kind=single),dimension(:),pointer :: Counts_Reb , Counts_Err_Reb
    integer,intent(inout) :: Size_Rebin_Vectors

    real,dimension(Size_Rebin_Vectors) :: Emed_Plot , Edelta_Plot
    real,dimension(Size_Rebin_Vectors) :: Counts_Plot , Counts_Err_Plot
    integer :: Size_Plot_Vectors , i
 
    Size_Plot_Vectors = 0
    do i=1,Size_Rebin_Vectors
      if ( (Emed_Reb(i)-Edelta_Reb(i)>0) .and. &
        (Counts_Reb(i)-Counts_Err_Reb(i)>0) ) then
        Size_Plot_Vectors = Size_Plot_Vectors + 1
        Emed_Plot(Size_Plot_Vectors)=Emed_Reb(i)
        Edelta_Plot(Size_Plot_Vectors)=Edelta_Reb(i) 
        Counts_Plot(Size_Plot_Vectors)=Counts_Reb(i) 
        Counts_Err_Plot(Size_Plot_Vectors)=Counts_Err_Reb(i) 
      end if
    end do
    deallocate(Emed_Reb , Edelta_Reb)
    deallocate(Counts_Reb , Counts_Err_Reb)
    allocate(Emed_Reb(Size_Plot_Vectors) , Edelta_Reb(Size_Plot_Vectors) )
    allocate(Counts_Reb(Size_Plot_Vectors) , Counts_Err_Reb(Size_Plot_Vectors) )
    Emed_Reb=Emed_Plot ; Edelta_Reb=Edelta_Plot 
    Counts_Reb=Counts_Plot ; Counts_Err_Reb=Counts_Err_Plot 
    if (Size_Plot_Vectors==0) call error('ESPECPLOT', &
         'No valid data points to plot on logarithmic scale')
    Size_Rebin_Vectors = Size_Plot_Vectors

 
  end subroutine eliminate_bad_values

  !-------------------------------------------------------------------------------------
  subroutine normalize_spectrum(Edelta_Plot,Counts_Plot,Counts_Err_Plot,Exposure_Time)
  !-------------------------------------------------------------------------------------
    real(kind=single),dimension(:),pointer ::  Edelta_Plot
    real(kind=single),dimension(:),pointer :: Counts_Plot , Counts_Err_Plot
    real,intent(in) :: Exposure_Time

    if ( minval(Edelta_Plot)<=epsilon(0.0) ) call error('ESPECPLOT', &
      'At least one energy bin has a width negative or null')
    Counts_Plot = Counts_Plot/(2. * (Edelta_Plot/1000.0) * Exposure_Time)
    Counts_Err_Plot = Counts_Err_Plot/(2. * (Edelta_Plot/1000.0) * Exposure_Time)

  end subroutine normalize_spectrum

  !-------------------------------------------------------------------------------------
  subroutine open_plot(Device_Name , Output_File , &
    Label_Gen , Label_Obs , Label_Exposure , Label_Time , Window1 , Window2)
  !-------------------------------------------------------------------------------------
    character(len=1024),intent(inout) :: Device_Name , Output_File
    character(len=1000),intent(in) :: Label_Gen, Label_Obs, Label_Exposure, Label_Time
    real,dimension(4),intent(out) :: Window1
    real,dimension(4),intent(out) :: Window2
    
    character(len=8),dimension(7) :: Device_Table = (/ "/XWINDOW" , "/PS     " , &
         "/CPS    " , "/VPS    " , "/VCPS   " , "/GIF    " , "/VGIF   " /)
    integer :: pgbeg , i
    character(len=2048) :: Device_Selected
    logical :: Good_Device , File_Exists

    Window1 = (/0.05,0.95,0.50,0.85/) 
    Window2 = (/0.05,0.95,0.05,0.40/)
    if (Device_Name(1:1) /= '/') Device_Name='/'//Device_Name
    Good_Device = .false.
    do i=1,7
      if ( trim(Device_Name) == trim(Device_Table(i)) ) Good_Device = .true.
    end do
    if (.not. Good_Device) call error('ESPECPLOT', &
         'The device type '//trim(Device_Name)//' is not recognized by the task')
    if ( trim(Device_Name) == '/XWINDOW' ) then
      Device_Selected=Device_Name
    else     
       
      Device_Selected=trim(Output_File)//trim(Device_Name)
      inquire(file=Output_File,exist=File_Exists)
      if (File_Exists .and. .not. clobber()) call error('ESPECPLOT', &
      'Already existing file '//trim(Output_File)//' cannot be overwritten (SAS_CLOBER=0)')
      if (File_Exists .and. clobber()) call warning('ESPECPLOT', &
      'Already existing file '//trim(Output_File)//' will be overwritten')
   end if
    
    if ( pgbeg(0,Device_Selected,1,1)/=1 ) &
      call error('ESPECPLOT','Invalid device '//trim(Device_Selected))
    call pgpap(7.5 , 1.0 )
    call pgask(.true.)
    call pgsvp(0.0,1.0,0.0,1.0)
    call pgsch(0.5)
    call pgsci(10)
    call pgmtxt('T' , -1. , 0.5 , 0.5 , Label_gen)
    call pgmtxt('T' , -2.5 , 0.5 , 0.5 , Label_obs)
    call pgmtxt('T' , -4. , 0.5 , 0.5 , Label_Time)
    call pgmtxt('T' , -5.5 , 0.5 , 0.5 , Label_exposure)
    call pgsci(10)

  end subroutine open_plot

  !-------------------------------------------------------------------------------------
  subroutine draw_plot(Window , Title , Size_Plot_Vectors , &
    Emed_Plot , Edelta_Plot , Counts_Plot , Counts_Err_Plot)
  !-------------------------------------------------------------------------------------
    real,dimension(4),intent(in) :: Window
    character(len=*) :: Title
    integer,intent(in) :: Size_Plot_Vectors
    real(kind=single),dimension(:),pointer :: Emed_Plot , Edelta_Plot
    real(kind=single),dimension(:),pointer :: Counts_Plot , Counts_Err_Plot

    call pgsvp(Window(1),Window(2),Window(3),Window(4))

    call pgswin(log10(0.1) , log10(10.0) , &
      minval(log10(Counts_Plot - Counts_Err_Plot)) , &
      maxval(log10(Counts_Plot + Counts_Err_Plot)) )

    call pgsci(10)
    call pgbox('BCLNST',0.0,0,'BCLNST',0.0,0)
    call pglab('Energy (keV)' , 'Counts s\u-1\d keV\u-1\d' , Title)
    
    call pgsci(10)
    call pgerrx(Size_Plot_Vectors , log10(Emed_Plot-Edelta_Plot) , &
      log10(Emed_Plot+Edelta_Plot) , log10(Counts_Plot) , 0.)
    call pgerry(Size_Plot_Vectors , log10(Emed_Plot/1000.0) , &
      log10(Counts_Plot-Counts_Err_Plot) , log10(Counts_Plot+Counts_Err_Plot) , 0.)
    call pgsci(10) ; call pgslw(10)
    call pgpt(Size_Plot_Vectors , log10(Emed_Plot/1000.0) , log10(Counts_Plot) , -1)
    call pgslw(1)


  end subroutine draw_plot

  !-------------------------------------------------------------------------------------
  function change_case(instring,dir) result(outstring)
  !-------------------------------------------------------------------------------------
  ! Switch from lower to upper case or vice-versa: From Jean Ballet
  ! instring : input string
  ! dir      : direction (1: to lower case; 2: to uppercase; 0: switch case)

    character(len=*), intent(in) :: instring
    integer, intent(in) :: dir
    character(len=len(instring)) :: outstring

  ! ASCII positions of lower case and upper case letters

    integer, parameter :: start_up  = iachar("A"), end_up  = iachar("Z"), &
                          start_low = iachar("a"), end_low = iachar("z"), &
                          offset = start_low - start_up

    integer :: length, icur, pos

    length = len(instring)
    outstring = instring
    do icur = 1, length
       pos = iachar(instring(icur:icur))
       if (pos >= start_up .and. pos <= end_up .and. dir < 2) then
          pos = pos + offset
       else if (pos >= start_low .and. pos <= end_low .and. dir /= 1) then
          pos = pos - offset
       endif
       outstring(icur:icur) = achar(pos)
    enddo

  end function change_case    

end module especplot_mod
